import uuid
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union, cast

from turnip_text import (
    Block,
    BlockScope,
    BlockScopeBuilder,
    Inline,
    InlineScope,
    InlineScopeBuilder,
    Paragraph,
    Sentence,
    UnescapedText,
)
from turnip_text.helpers import block_scope_builder, inline_scope_builder
from turnip_text.renderers import CustomRenderFunc, Renderer, RendererPlugin
from turnip_text.renderers.dictify import dictify_pure_property
from turnip_text.renderers.std_plugins import (
    CitationPluginInterface,
    CiteKey,
    FootnotePluginInterface,
    FormatPluginInterface,
    SectionPluginInterface,
)

from .base import RawLatex

CiteKeyWithOptionNote = Tuple[CiteKey, Optional[UnescapedText]]


@dataclass(frozen=True)
class FootnoteAnchor(Inline):
    label: str


@dataclass(frozen=True)
class HeadedBlock(Block):
    latex_name: str
    name: UnescapedText
    contents: BlockScope
    label: Optional[str] = None
    num: bool = True


@dataclass(frozen=True)
class Citation(Inline):
    # List of (label, note?)
    labels: List[CiteKeyWithOptionNote]


@dataclass(frozen=True)
class NamedUrl(Inline, InlineScopeBuilder):
    url: str
    name: Optional[InlineScope] = None

    def build_from_inlines(self, inls: InlineScope) -> Inline:
        assert self.name is None
        return NamedUrl(self.url, inls)


@dataclass(frozen=True)
class DisplayList(Block):
    # TODO allow nested lists
    # items: List[Union[BlockNode, List]]
    items: List["DisplayListItem"]
    mode: str


@dataclass(frozen=True)
class DisplayListItem(Block):
    item: Block


@dataclass(frozen=True)
class Formatted(Inline):
    format_type: str  # e.g. "emph"
    items: InlineScope


class LatexCitationPlugin(RendererPlugin, CitationPluginInterface):
    # TODO require biblatex

    _citations: Dict[str, Any]

    def __init__(self) -> None:
        super().__init__()

        # TODO load citations from somewhere
        self._citations = {}

    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((Citation, self._render_citation),)

    def _postamble_handlers(self) -> Iterable[Tuple[str, Callable[[Renderer], str]]]:
        return ((self._BIBLIOGRAPHY_POSTAMBLE_ID, self._render_bibliography),)

    def _render_citation(self, renderer: Renderer, citation: Citation) -> str:
        if any(note for _, note in citation.labels):
            # We can't add a citenote for multiple citations at a time in Latex - split individually
            data = ""
            for key, note in citation.labels:
                if note:
                    rendered_note = renderer.render_unescapedtext(note)
                    data += f"\\cite[{rendered_note}]{{{key}}}"
                else:
                    data += f"\\cite{{{key}}}"
            return data
        else:
            data = "\\cite{" + ",".join(key for key, _ in citation.labels) + "}"
            return data

    def _render_bibliography(self, renderer: Renderer) -> str:
        return """{
\\raggedright
\\printbibliography
}"""

    def cite(self, *labels: Union[str, Tuple[str, str]]) -> Inline:
        # Convert ["label"] to [("label", None)] so Citation has a consistent format
        adapted_labels = [
            (label, None)
            if isinstance(label, str)
            else (label[0], UnescapedText(label[1]))
            for label in labels
        ]
        return Citation(adapted_labels)

    # TODO make this output \citeauthor
    def citeauthor(self, label: str) -> Inline:
        return Citation([(label, None)])


class LatexFootnotePlugin(RendererPlugin, FootnotePluginInterface):
    _footnotes: Dict[str, Block]

    def __init__(self) -> None:
        super().__init__()

        self._footnotes = {}

    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((FootnoteAnchor, self._render_footnote_anchor),)

    def _render_footnote_anchor(
        self, renderer: Renderer, footnote: FootnoteAnchor
    ) -> str:
        # TODO - intelligent footnotetext placement using floats?
        rendered_footnotetext = renderer.render_block(self._footnotes[footnote.label])
        return f"\\footnote{{{rendered_footnotetext}}}"

    @dictify_pure_property
    def footnote(self) -> InlineScopeBuilder:
        @inline_scope_builder
        def footnote_builder(contents: InlineScope) -> Inline:
            label = str(uuid.uuid4())
            self._footnotes[label] = Paragraph([Sentence([contents])])
            return FootnoteAnchor(label)

        return footnote_builder

    def footnote_ref(self, label: str) -> Inline:
        return FootnoteAnchor(label)

    def footnote_text(self, label: str) -> BlockScopeBuilder:
        # Return a callable which is invoked with the contents of the following inline scope
        # Example usage:
        # [footnote_text("label")]{text}
        # equivalent to
        # [footnote_text("label")(r"text")]
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Optional[Block]:
            self._footnotes[label] = contents
            return None

        return handle_block_contents


class LatexSectionPlugin(RendererPlugin, SectionPluginInterface):
    _pagebreak_before: List[str]

    def __init__(self, pagebreak_before: List[str] = []) -> None:
        super().__init__()
        self._pagebreak_before = pagebreak_before

    def _block_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((HeadedBlock, self._render_headed_block),)

    def _render_headed_block(self, renderer: Renderer, block: HeadedBlock) -> str:
        header = f"\\{block.latex_name}"  # i.e. r"\section"
        if block.latex_name in self._pagebreak_before:
            header = "\\pagebreak\n" + header
        if not block.num:
            header += "*"
        escaped_name = renderer.render_unescapedtext(block.name)
        header += f"{{{escaped_name}}}"  # i.e. r"\section*" + "{Section Name}"
        if block.label:
            header += f"\\label{{{block.label}}}"  # i.e. r"\section*{Section Name}" + r"\label{block_label}"
        return f"{header}\n\n" + renderer.render_blockscope(block.contents)

    def section(
        self, name: str, label: Optional[str] = None, num: bool = True
    ) -> BlockScopeBuilder:
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                latex_name="section",
                name=UnescapedText(name),
                label=label,
                num=num,
                contents=contents,
            )

        return handle_block_contents

    def subsection(
        self, name: str, label: Optional[str] = None, num: bool = True
    ) -> BlockScopeBuilder:
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                latex_name="subsection",
                name=UnescapedText(name),
                label=label,
                num=num,
                contents=contents,
            )

        return handle_block_contents

    def subsubsection(
        self, name: str, label: Optional[str] = None, num: bool = True
    ) -> BlockScopeBuilder:
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                latex_name="subsubsection",
                name=UnescapedText(name),
                label=label,
                num=num,
                contents=contents,
            )

        return handle_block_contents


@inline_scope_builder
def emph_builder(items: InlineScope) -> Inline:
    return Formatted("emph", items)


@inline_scope_builder
def italic_builder(items: InlineScope) -> Inline:
    return Formatted("textit", items)


@inline_scope_builder
def bold_builder(items: InlineScope) -> Inline:
    return Formatted("textbf", items)


class LatexFormatPlugin(RendererPlugin, FormatPluginInterface):
    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((Formatted, self._render_formatted),)

    def _render_formatted(self, renderer: Renderer, item: Formatted) -> str:
        data = f"\\{item.format_type}{{"
        data += renderer.render_inlinescope(item.items)
        return data + "}"

    emph = emph_builder
    italic = italic_builder
    bold = bold_builder

    OPEN_DQUOTE = RawLatex("``")
    CLOS_DQUOTE = RawLatex("''")

    @dictify_pure_property
    def enquote(self) -> InlineScopeBuilder:
        @inline_scope_builder
        def enquote_builder(items: InlineScope) -> Inline:
            return InlineScope(
                [LatexFormatPlugin.OPEN_DQUOTE]
                + list(items)
                + [LatexFormatPlugin.CLOS_DQUOTE]
            )

        return enquote_builder


class LatexListPlugin(RendererPlugin):
    def _block_handlers(self) -> Iterable[CustomRenderFunc]:
        return (
            (DisplayList, self._render_list),
            (DisplayListItem, self._render_list_item),
        )

    def _render_list(self, renderer: Renderer, list: DisplayList) -> str:
        # TODO indents!
        data = f"\\begin{{{list.mode}}}\n"
        data += renderer.PARAGRAPH_SEP.join(
            renderer.render_block(i) for i in list.items
        )
        return data + f"\n\\end{{{list.mode}}}"

    def _render_list_item(self, renderer: Renderer, list_item: DisplayListItem) -> str:
        # TODO indents!
        # Put {} after \item so square brackets at the start of render_block don't get swallowed as arguments
        return "\\item{} " + renderer.render_block(list_item.item)

    @dictify_pure_property
    def enumerate(self) -> BlockScopeBuilder:
        @block_scope_builder
        def enumerate_builder(contents: BlockScope) -> Block:
            items = list(contents)
            if any(not isinstance(x, DisplayListItem) for x in items):
                raise TypeError(
                    f"Found blocks in this list that were not list [item]s!"
                )
            return DisplayList(
                mode="enumerate", items=cast(List[DisplayListItem], items)
            )

        return enumerate_builder

    @dictify_pure_property
    def itemize(self) -> BlockScopeBuilder:
        @block_scope_builder
        def itemize_builder(contents: BlockScope) -> Block:
            items = list(contents)
            if any(not isinstance(x, DisplayListItem) for x in items):
                raise TypeError(
                    f"Found blocks in this list that were not list [item]s!"
                )
            return DisplayList(mode="itemize", items=cast(List[DisplayListItem], items))

        return itemize_builder

    @dictify_pure_property
    def item(self) -> BlockScopeBuilder:
        @block_scope_builder
        def item_builder(block_scope: BlockScope) -> Block:
            return DisplayListItem(block_scope)

        return item_builder


class LatexUrlPlugin(RendererPlugin):
    # TODO add dependency on hyperref!!

    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((NamedUrl, self._render_url),)

    def _render_url(self, renderer: Renderer, url: NamedUrl) -> str:
        if url.name is None:
            return f"\\url{{{url.url}}}"
        else:
            escaped_url_name = renderer.render_inlinescope(url.name)
            return f"\\href{{{url.url}}}{{{escaped_url_name}}}"

    def url(self, url: str, name: Optional[str] = None) -> Inline:
        return NamedUrl(
            url, name=InlineScope([UnescapedText(name)]) if name is not None else None
        )



from typing import Optional, Protocol, Union, runtime_checkable

from ._native import (  # type: ignore
    BlockScope,
    InlineScope,
    Paragraph,
    RawText,
    Sentence,
    UnescapedText,
)
from ._native import parse_file as parse_file_native  # type: ignore
from ._native import parse_str as parse_str_native  # type: ignore


class Inline(Protocol):
    is_inline: bool = True


@runtime_checkable
class Block(Protocol):
    is_block: bool = True


@runtime_checkable
class BlockScopeBuilder(Protocol):
    def build_from_blocks(self, bs: BlockScope) -> Optional[Block]:
        ...


@runtime_checkable
class InlineScopeBuilder(Protocol):
    def build_from_inlines(self, inls: InlineScope) -> Inline:
        ...


@runtime_checkable
class RawScopeBuilder(Protocol):
    def build_from_raw(self, raw: str) -> Union[Inline, Block]:
        ...



import abc
from os import PathLike
from typing import (
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Tuple,
    Type,
    TypeVar,
)

from turnip_text import (
    Block,
    BlockScope,
    Inline,
    InlineScope,
    Paragraph,
    Sentence,
    UnescapedText,
    parse_file_native,
)
from turnip_text.renderers.dictify import dictify

T = TypeVar("T")
CustomRenderFunc = Tuple[Type[T], Callable[["Renderer", T], str]]


class TypeToRenderMap(Generic[T]):
    _handlers: Dict[Type[T], Callable[["Renderer", T], str]]

    def __init__(self) -> None:
        super().__init__()
        self._handlers = {}

    def push_association(self, t_func: CustomRenderFunc):
        t, func = t_func
        if t in self._handlers:
            raise RuntimeError(f"Conflict: registered two renderers for {t}")
        self._handlers[t] = func

    def render(self, base_renderer: "Renderer", obj: T) -> str:
        for t, renderer in self._handlers.items():
            if isinstance(obj, t):
                return renderer(base_renderer, obj)
        raise NotImplementedError(f"Couldn't handle {obj}")


# TODO Make preamble/postamble return Blocks to be rendered instead of just str? Would allow e.g. a Bibliography section? Perhaps better to expose a bibliography block for "standard" postambles?
class AmbleMap:
    """Class that stores and reorders {pre,post}amble handlers.

    Handlers are simply functions that return a string, which have a unique ID.

    Only one handler per ID can exist.

    The user can request that the ID are sorted in a particular order, or default to order of insertion.

    When the document is rendered, the handlers will be called in that order."""

    _handlers: Dict[str, Callable[["Renderer"], str]]
    _id_order: List[str]

    def __init__(self) -> None:
        self._handlers = {}
        self._id_order = []

    def push_handler(self, id: str, f: Callable[["Renderer"], str]):
        if id in self._handlers:
            raise RuntimeError(f"Conflict: registered two amble-handlers for ID {id}")
        self._handlers[id] = f
        self._id_order.append(id)

    def reorder_handlers(self, selected_id_order: List[str]):
        """Request that certain handler IDs are rendered in a specific order.

        Does not need to be a complete ordering, i.e. if handlers ['a', 'b', 'c'] are registered
        this function can be called with ['c', 'a'] to ensure 'c' comes before 'a',
        but all of the IDs in the order need to have been registered.

        When the requested ordering is incomplete, handlers which haven't been mentioned
        will retain their old order but there is no specified ordering between (mentioned) and (not-mentioned) IDs.
        """

        assert all(id in self._handlers for id in selected_id_order)

        if len(selected_id_order) != len(set(selected_id_order)):
            raise RuntimeError(
                f"reorder_handlers() called with ordering with duplicate IDs: {selected_id_order}"
            )

        # Shortcut if the selected order is complete i.e. covers all IDs so far
        if len(self._id_order) == len(selected_id_order):
            self._id_order = selected_id_order
        else:
            # Otherwise, we need to consider the non-selected IDs too.
            # The easy way: put selected ones first, then non-selected ones last
            # Get the list of ids NOT in selected_id_order, in the order they're currently in in self._id_order
            non_selected_ids = [
                id for id in self._id_order if id not in selected_id_order
            ]
            self._id_order = selected_id_order + non_selected_ids

        assert all(id in self._id_order for id in self._handlers.keys())

    def generate_ambles(self, renderer: "Renderer") -> Iterator[str]:
        for id in self._id_order:
            yield self._handlers[id](renderer)


class Renderer(abc.ABC):
    PARAGRAPH_SEP: str
    SENTENCE_SEP: str

    plugins: List["RendererPlugin"]

    block_handlers: TypeToRenderMap[Block]
    inline_handlers: TypeToRenderMap[Inline]
    preamble_handlers: AmbleMap
    postamble_handlers: AmbleMap

    def __init__(self, plugins: List["RendererPlugin"]) -> None:
        super().__init__()

        # Create handlers and
        self.block_handlers = TypeToRenderMap()
        self.block_handlers.push_association(
            (BlockScope, lambda _, bs: self.render_blockscope(bs))
        )
        self.block_handlers.push_association(
            (Paragraph, lambda _, bs: self.render_paragraph(bs))
        )

        self.inline_handlers = TypeToRenderMap()
        self.inline_handlers.push_association(
            (InlineScope, lambda _, inls: self.render_inlinescope(inls))
        )
        self.inline_handlers.push_association(
            (UnescapedText, lambda _, t: self.render_unescapedtext(t))
        )

        self.preamble_handlers = AmbleMap()
        self.postamble_handlers = AmbleMap()

        self.plugins = plugins
        for p in self.plugins:
            for block_t_func in p._block_handlers():
                self.block_handlers.push_association(block_t_func)
            for inl_t_func in p._inline_handlers():
                self.inline_handlers.push_association(inl_t_func)
            for preamble_id, preamble_func in p._preamble_handlers():
                self.preamble_handlers.push_handler(preamble_id, preamble_func)
            for postamble_id, postamble_func in p._postamble_handlers():
                self.postamble_handlers.push_handler(postamble_id, postamble_func)

    def request_preamble_order(self, preamble_id_order: List[str]):
        self.preamble_handlers.reorder_handlers(preamble_id_order)

    def request_postamble_order(self, postamble_id_order: List[str]):
        self.postamble_handlers.reorder_handlers(postamble_id_order)

    def parse_file(self, p: PathLike) -> BlockScope:
        # TODO this seems super icky
        # The problem: we want to be able to call e.g. [footnote] inside the turniptext file.
        # But if footnote were a free function, it would mutate global state -- we don't want that.
        # A hack! Require a 'renderer object' to be passed in - this encapsulates the local state.
        # Create a new dictionary, holding all of the Renderer's public fields,
        # and use that as the locals for parse_file_local.

        locals = {}
        for plugin in self.plugins:
            locals.update(dictify(plugin))

        return parse_file_native(str(p), locals)

    def render_unescapedtext(self, t: UnescapedText) -> str:
        """The baseline - take text and return a string that will look like that text exactly in the given backend."""
        raise NotImplementedError(f"Need to implement render_unescapedtext")

    def render_doc(self, doc_block: BlockScope) -> str:
        doc = ""
        for preamble in self.preamble_handlers.generate_ambles(self):
            doc += preamble
            doc += self.PARAGRAPH_SEP
        doc += self.render_blockscope(doc_block)
        for postamble in self.postamble_handlers.generate_ambles(self):
            doc += self.PARAGRAPH_SEP
            doc += postamble
        return doc

    def render_inline(self, i: Inline) -> str:
        return self.inline_handlers.render(self, i)

    def render_block(self, b: Block) -> str:
        return self.block_handlers.render(self, b)

    def render_blockscope(self, bs: BlockScope) -> str:
        # Default: join paragraphs with self.PARAGRAPH_SEP
        # If you get nested blockscopes, this will still be fine - you won't get double separators
        return self.PARAGRAPH_SEP.join(self.render_block(b) for b in bs)

    def render_paragraph(self, p: Paragraph) -> str:
        # Default: join sentences with self.SENTENCE_SEP
        return self.SENTENCE_SEP.join(self.render_sentence(s) for s in p)

    def render_inlinescope(self, inls: InlineScope) -> str:
        # Default: join internal inline elements directly
        return "".join(self.render_inline(i) for i in inls)

    def render_sentence(self, s: Sentence) -> str:
        # Default: join internal inline elements directly
        # TODO could be extended by e.g. latex to ensure you get sentence-break-whitespace at the end of each sentence?
        return "".join(self.render_inline(i) for i in s)


class RendererPlugin(abc.ABC):
    """
    Plugins should export a consistent set of functions for documents to rely on.

    These should only use 'pure' properties, that don't mutate state, because they won't be called multiple times - see `dictify_renderer()` function for details why.
    'Pure' properties need to be marked with `@dictify_pure_property`.

    If you want impure results (i.e. once that mutate state) without function call syntax, you can use Builders.
    e.g. the `footnote` property is pure, but returns a single builder object that mutates the internal list of footnotes whenever `.build()` is called.
    """

    @property
    def _plugin_name(self) -> str:
        return type(self).__name__

    def _block_handlers(self) -> Iterable[CustomRenderFunc[Block]]:
        return ()

    def _inline_handlers(self) -> Iterable[CustomRenderFunc[Inline]]:
        return ()

    def _preamble_handlers(self) -> Iterable[Tuple[str, Callable[[Renderer], str]]]:
        return ()

    def _postamble_handlers(self) -> Iterable[Tuple[str, Callable[[Renderer], str]]]:
        return ()



from dataclasses import dataclass
from typing import List

from turnip_text import Inline, UnescapedText
from turnip_text.renderers import Renderer, RendererPlugin


@dataclass(frozen=True)
class RawMarkdown(Inline):
    md: str


class MarkdownRenderer(Renderer):
    PARAGRAPH_SEP = "\n\n"
    SENTENCE_SEP = "\n"

    def __init__(self, plugins: List["RendererPlugin"]) -> None:
        super().__init__(plugins)
        self.inline_handlers.push_association(
            (RawMarkdown, lambda _, raw: self._render_raw_markdown(raw))
        )

    def _render_raw_markdown(self, r: RawMarkdown) -> str:
        return r.md

    def render_unescapedtext(self, t: UnescapedText) -> str:
        # note - right now this assumes we're using a unicode-compatible setup and thus don't need to escape unicode characters.
        # note - order is important here because the subsitutions may introduce more special characters.
        # e.g. if the backslash replacement applied later, the backslash from escaping "(" would be escaped as well

        # TODO - some of these are overzealous, e.g. () and -, because in most contexts they're interpreted as normal text.
        # context sensitive escaping?
        # https://www.markdownguide.org/basic-syntax/
        special_chars = r"\\`*_{}[]<>()#+-.!|"

        data = t.text
        for char in special_chars:
            data = data.replace(char, "\\" + char)
        return data



from typing import Any, Dict, List, Mapping


class dictify_pure_property(property):
    """Equivalent to `property`, but acts as a purity marker.
    This should only be used if invoking the property itself DOESN'T MUTATE STATE.
    This means calling it once is equivalent to calling it many times, which is a useful property when dictifying."""
    pass

def dictify(r: Any) -> Dict[str, Any]:
    """
    Given an object implementing `Renderer`, get a dict of all functions, methods, and fields it exposes publically.

    Public = does not begin with '_'. This hides internal Python methods (e.g. `__str__`), name-mangled variables (both the original var `__name_mangled` and the mangled `_ExampleClass__name_mangled`, and any variables the programmer doesn't wish to expose `_plz_dont_modify_directly`.

    These are retrieved as follows:
    1. Get `dir(r)`, to get a dictionary of the fields and descriptors it exposes.
    2. Iterate through the keys, filtering out any that begin with `_` to find the public fields.
    3. use `getattr(r, key)` to get the values of those fields and descriptors, (which could be bound methods, static methods, or plain values), putting them into a new dictionary, which is returned.
        a. Warn the user if `type(r).__dict__[key]` is an impure DATA DESCRIPTOR, e.g. a property. These are evaluated ONCE inside this function, and wouldn't be repeatedly evaluated when using the returned dict.
        That is different to the usual behaviour: if `key` is a property, reading `r.key` will call `type(r).__dict__[key].__get__(...)` every time. `returned_dict[key]` holds the value returned from calling that ONCE, and repeatedly reading it will not re-invoke the property getter.
        IF THE PROPERTY IS PURE you can avoid this warning by using @dictify_pure_property to declare it.

    This can be used as the execution environment for code inside a turnip_text file.

    [1]: Information on Python "descriptors" https://docs.python.org/3.8/howto/descriptor.html
    """

    from inspect import isdatadescriptor

    r_obj_public_fields: List[str] = [
        k
        for k in dir(r)
        if not k.startswith("_")
    ]
    
    r_type_dict: Mapping[str, Any] = type(r).__dict__
    
    # Warn about impure data descriptor fields
    for k in r_obj_public_fields:
        if isdatadescriptor(r_type_dict[k]) and not isinstance(r_type_dict[k], dictify_pure_property):
            print(f"dictify_renderer Warning: renderer {r} exposes a public 'data descriptor' (e.g. a property) "
                  f"named {k!r}. This will be evaluated exactly once, and the result will be stored in the "
                  f"returned dict, instead of evaluating the property each time the dict is accessed. "
                  f"DO NOT USE THESE IF YOU CAN AVOID IT. Use a normal field instead.")

    return {
        k: getattr(r, k)
        for k in r_obj_public_fields
    }


import abc
from typing import Iterator, List, Optional, Protocol, runtime_checkable

from turnip_text._native import Block, Inline

class UnescapedText(Inline):
    def __init__(self, text: str) -> None: ...
    @property
    def text(self) -> str: ...

class RawText(Inline):
    def __init__(self, text: str) -> None: ...
    @property
    def text(self) -> str: ...

# Note - Sentence is NOT an Inline. This means there's always a hierarchy of Paragraph -> many Sentences -> many Inlines.
# InlineScopes can be nested, Sentences cannot.
class Sentence:
    def __init__(self, list: Optional[List[Inline]] = None): ...
    def __len__(self) -> int: ...
    # Iterate over the inline blocks in the sentence
    def __iter__(self) -> Iterator: ...
    # Push an inline node into the sentence
    def push_inline(self, node: Inline): ...

class Paragraph(Block):
    def __init__(self, list: Optional[List[Sentence]] = None): ...
    def __len__(self) -> int: ...
    # Iterate over the sentences in the Paragraph
    def __iter__(self) -> Iterator[Sentence]: ...
    # Push a sentence into the Paragraph
    def push_sentence(self, s: Sentence): ...

class BlockScope(Block):
    def __init__(self, list: Optional[List[Block]] = None): ...
    def __len__(self) -> int: ...
    # Iterate over the blocks in the BlockScope
    def __iter__(self) -> Iterator[Block]: ...
    # Push a block into the BlockScope
    def push_block(self, b: Block): ...

class InlineScope(Inline):
    def __init__(self, list: Optional[List[Inline]] = None): ...
    def __len__(self) -> int: ...
    # Iterate over the inline items in the InlineScope
    def __iter__(self) -> Iterator[Inline]: ...
    # Push an inline item into the InlineScope
    def push_inline(self, b: Inline): ...



import uuid
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from turnip_text import (
    Block,
    BlockScope,
    BlockScopeBuilder,
    Inline,
    InlineScope,
    InlineScopeBuilder,
    Paragraph,
    Sentence,
    UnescapedText,
)
from turnip_text.helpers import block_scope_builder, inline_scope_builder
from turnip_text.renderers import CustomRenderFunc, Renderer, RendererPlugin
from turnip_text.renderers.dictify import dictify_pure_property
from turnip_text.renderers.markdown.base import RawMarkdown
from turnip_text.renderers.std_plugins import (
    CitationPluginInterface,
    CiteKey,
    FootnotePluginInterface,
    FormatPluginInterface,
    SectionPluginInterface,
)

CiteKeyWithOptionNote = Tuple[CiteKey, Optional[UnescapedText]]


@dataclass(frozen=True)
class FootnoteAnchor(Inline):
    label: str


@dataclass(frozen=True)
class HeadedBlock(Block):
    level: int
    name: UnescapedText
    contents: BlockScope
    # Numbering
    use_num: bool  # Whether to number the heading or not
    num: Tuple[int, ...]  # The actual number of the heading
    # Labelling
    label: Optional[str] = None


@dataclass(frozen=True)
class Citation(Inline):
    # List of (label, note?)
    labels: List[CiteKeyWithOptionNote]


@dataclass(frozen=True)
class NamedUrl(Inline, InlineScopeBuilder):
    url: str
    name: Optional[InlineScope] = None

    def build_from_inlines(self, inls: InlineScope) -> Inline:
        assert self.name is None
        return NamedUrl(self.url, inls)


@dataclass(frozen=True)
class DisplayList(Block):
    # TODO allow nested lists
    # items: List[Union[BlockNode, List]]
    items: List["DisplayListItem"]
    numbered: bool


@dataclass(frozen=True)
class DisplayListItem(Block):
    contents: Block


@dataclass(frozen=True)
class Formatted(Inline):
    format_type: str  # e.g. "*", "**"
    items: InlineScope


class MarkdownCitationAsFootnotePlugin(RendererPlugin, CitationPluginInterface):
    _citations: Dict[str, Any]
    _referenced_citations: Set[str]

    def __init__(self) -> None:
        super().__init__()

        # TODO load citations from somewhere
        self._citations = {}
        self._referenced_citations = set()

    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((Citation, self._render_citation),)

    def _postamble_handlers(self) -> Iterable[Tuple[str, Callable[[Renderer], str]]]:
        return ((self._BIBLIOGRAPHY_POSTAMBLE_ID, self._render_bibliography),)

    def _render_citation(self, renderer: Renderer, citation: Citation) -> str:
        # TODO what happens with unmarkdownable labels? e.g. labels with backslash or something. need to check that when loading.
        # TODO also maybe people wouldn't want those labels being exposed?
        return "".join(
            f"[^{label}]"
            if opt_note is None
            else f"\\([^{label}], {renderer.render_unescapedtext(opt_note)}\\)"
            for label, opt_note in citation.labels
        )

    def _render_bibliography(self, renderer: Renderer) -> str:
        # TODO actual reference rendering!
        return renderer.PARAGRAPH_SEP.join(
            f"[^{label}]: cite {label}" for label in self._referenced_citations
        )

    def cite(self, *labels: Union[str, Tuple[str, str]]) -> Inline:
        # Convert ["label"] to [("label", None)] so Citation has a consistent format
        adapted_labels = [
            (label, None)
            if isinstance(label, str)
            else (label[0], UnescapedText(label[1]))
            for label in labels
        ]

        self._referenced_citations.update(label for label, _ in adapted_labels)

        return Citation(adapted_labels)

    # TODO make this output \citeauthor
    def citeauthor(self, label: str) -> Inline:
        return Citation([(label, None)])


class MarkdownCitationAsHTMLPlugin(RendererPlugin, CitationPluginInterface):
    _citations: Dict[str, Any]
    _referenced_citations: Set[str]

    def __init__(self) -> None:
        super().__init__()

        # TODO load citations from somewhere
        self._citations = {}
        self._referenced_citations = set()

    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((Citation, self._render_citation),)

    def _postamble_handlers(self) -> Iterable[Tuple[str, Callable[[Renderer], str]]]:
        return ((self._BIBLIOGRAPHY_POSTAMBLE_ID, self._render_bibliography),)

    def _get_citation_shorthand(
        self, renderer: Renderer, label: str, note: Optional[UnescapedText]
    ) -> UnescapedText:
        # TODO could do e.g. numbering here
        if note:
            return UnescapedText(f"[{label}, {note.text}]")
        return UnescapedText(f"[{label}]")

    def _render_citation(self, renderer: Renderer, citation: Citation) -> str:
        # TODO what happens with unmarkdownable labels? e.g. labels with backslash or something. need to check that when loading.
        # TODO also maybe people wouldn't want those labels being exposed?
        return "".join(
            # f'<a href="#cite-{label}">{self._get_citation_shorthand(renderer, label, opt_note)}</a>'
            f"[{renderer.render_unescapedtext(self._get_citation_shorthand(renderer, label, opt_note))}](#cite-{label})"
            for label, opt_note in citation.labels
        )

    def _render_bibliography(self, renderer: Renderer) -> str:
        # TODO actual reference rendering!
        return renderer.PARAGRAPH_SEP.join(
            f'<a id="cite-{label}">{renderer.render_unescapedtext(self._get_citation_shorthand(renderer, label, None))}: cite {label}</a>'
            for label in self._referenced_citations
        )

    def cite(self, *labels: Union[str, Tuple[str, str]]) -> Inline:
        # Convert ["label"] to [("label", None)] so Citation has a consistent format
        adapted_labels = [
            (label, None)
            if isinstance(label, str)
            else (label[0], UnescapedText(label[1]))
            for label in labels
        ]

        self._referenced_citations.update(label for label, _ in adapted_labels)

        return Citation(adapted_labels)

    # TODO make this output \citeauthor
    def citeauthor(self, label: str) -> Inline:
        return Citation([(label, None)])


class MarkdownFootnotePlugin(RendererPlugin, FootnotePluginInterface):
    _footnotes: Dict[str, Block]
    _footnote_ref_order: List[str]

    _MARKDOWN_FOOTNOTE_POSTAMBLE_ID = "MarkdownFootnotePlugin_Footnotes"

    def __init__(self) -> None:
        super().__init__()

        self._footnotes = {}
        self._footnote_ref_order = []

    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((FootnoteAnchor, self._render_footnote_anchor),)

    def _postamble_handlers(self) -> Iterable[Tuple[str, Callable[[Renderer], str]]]:
        return ((self._MARKDOWN_FOOTNOTE_POSTAMBLE_ID, self._render_footnotes),)

    def _render_footnote_anchor(
        self, renderer: Renderer, footnote: FootnoteAnchor
    ) -> str:
        footnote_num = self._footnote_ref_order.index(footnote.label)
        return f"[^{footnote_num + 1}]"

    def _render_footnotes(self, renderer: Renderer) -> str:
        return renderer.PARAGRAPH_SEP.join(
            # TODO render with indent
            f"[^{num + 1}]: " + renderer.render_block(self._footnotes[label])
            for num, label in enumerate(self._footnote_ref_order)
        )

    def _add_footnote_reference(self, label: str):
        if label not in self._footnote_ref_order:
            self._footnote_ref_order.append(label)

    @dictify_pure_property
    def footnote(self) -> InlineScopeBuilder:
        @inline_scope_builder
        def footnote_builder(contents: InlineScope) -> Inline:
            label = str(uuid.uuid4())
            self._add_footnote_reference(label)
            self._footnotes[label] = Paragraph([Sentence([contents])])
            return FootnoteAnchor(label)

        return footnote_builder

    def footnote_ref(self, label: str) -> Inline:
        self._add_footnote_reference(label)
        return FootnoteAnchor(label)

    def footnote_text(self, label: str) -> BlockScopeBuilder:
        # Return a callable which is invoked with the contents of the following inline scope
        # Example usage:
        # [footnote_text("label")]{text}
        # equivalent to
        # [footnote_text("label")(r"text")]
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Optional[Block]:
            self._footnotes[label] = contents
            return None

        return handle_block_contents


class MarkdownSectionPlugin(RendererPlugin, SectionPluginInterface):
    def _block_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((HeadedBlock, self._render_headed_block),)

    def _render_headed_block(self, renderer: Renderer, block: HeadedBlock) -> str:
        if block.label:
            raise NotImplementedError("Section labelling not supported")

        header = "#" * block.level + " "
        if block.use_num:
            header += ".".join(str(n) for n in block.num)
            header += " "
        header += renderer.render_unescapedtext(block.name)

        return (
            header + renderer.PARAGRAPH_SEP + renderer.render_blockscope(block.contents)
        )

    # [section_num, subsection_num... etc]
    _current_number: List[int]

    def __init__(self) -> None:
        super().__init__()
        self._current_number = []

    def _update_number(self, current_level: int) -> Tuple[int, ...]:
        """Update the structure number based on a newly-encountered header at level current_level.

        i.e. if encountering a new section, call _update_number(2), which will a) increment the subsection number and b) return a tuple of (section_num, subsection_num) which can be used in rendering later.
        """

        # Step 1: pad self._current_number out to current_level elements with 0s
        self._current_number += (current_level - len(self._current_number)) * [0]
        # Step 2: increment self._current_number[level - 1] and reset all numbers beyond that to 0
        # e.g. for a plain section, level = 1 => increment current_number[0] and zero out current_number[1:]
        self._current_number[current_level - 1] += 1
        for i in range(current_level, len(self._current_number)):
            self._current_number[i] = 0
        # Step 3: return a tuple of all elements up to current_level
        return tuple(self._current_number[:current_level])

    def section(
        self, name: str, label: Optional[str] = None, num: bool = True
    ) -> BlockScopeBuilder:
        # Increment the section number here, not at the end of the block
        section_num = self._update_number(1)

        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                level=1,
                name=UnescapedText(name),
                label=label,
                use_num=num,
                num=section_num,
                contents=contents,
            )

        return handle_block_contents

    def subsection(
        self, name: str, label: Optional[str] = None, num: bool = True
    ) -> BlockScopeBuilder:
        # Increment the section number here, not at the end of the block
        section_num = self._update_number(2)

        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                level=2,
                name=UnescapedText(name),
                label=label,
                use_num=num,
                num=section_num,
                contents=contents,
            )

        return handle_block_contents

    def subsubsection(
        self, name: str, label: Optional[str] = None, num: bool = True
    ) -> BlockScopeBuilder:
        # Increment the section number here, not at the end of the block
        section_num = self._update_number(3)

        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                level=3,
                name=UnescapedText(name),
                label=label,
                use_num=num,
                num=section_num,
                contents=contents,
            )

        return handle_block_contents


@inline_scope_builder
def emph_builder(items: InlineScope) -> Inline:
    return italic_builder.build_from_inlines(items)


@inline_scope_builder
def italic_builder(items: InlineScope) -> Inline:
    # Use single underscore for italics, double asterisks for bold, double underscore for underlining?
    return Formatted("_", items)


@inline_scope_builder
def bold_builder(items: InlineScope) -> Inline:
    return Formatted("**", items)


class MarkdownFormatPlugin(RendererPlugin, FormatPluginInterface):
    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((Formatted, self._render_formatted),)

    def _render_formatted(self, renderer: Renderer, item: Formatted) -> str:
        return (
            item.format_type
            + renderer.render_inlinescope(item.items)
            + item.format_type
        )

    emph = emph_builder
    italic = italic_builder
    bold = bold_builder

    DQUOTE = RawMarkdown('"')

    @dictify_pure_property
    def enquote(self) -> InlineScopeBuilder:
        @inline_scope_builder
        def enquote_builder(items: InlineScope) -> Inline:
            return InlineScope(
                [MarkdownFormatPlugin.DQUOTE]
                + list(items)
                + [MarkdownFormatPlugin.DQUOTE]
            )

        return enquote_builder


def indent_item(prefix, item_rendering):
    return prefix + item_rendering.replace("\n", "\n" + " " * len(prefix))


class MarkdownListPlugin(RendererPlugin):
    def _block_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((DisplayList, self._render_list),)

    def _render_list(self, renderer: Renderer, list: DisplayList) -> str:
        # TODO better way of indenting
        if list.numbered:
            return renderer.SENTENCE_SEP.join(
                indent_item(f"{idx+1}. ", renderer.render_block(item.contents))
                for idx, item in enumerate(list.items)
            )
        else:
            return renderer.SENTENCE_SEP.join(
                indent_item("- ", renderer.render_block(item.contents))
                for idx, item in enumerate(list.items)
            )

    @dictify_pure_property
    def enumerate(self) -> BlockScopeBuilder:
        @block_scope_builder
        def enumerate_builder(contents: BlockScope) -> Block:
            items = list(contents)
            if any(not isinstance(x, DisplayListItem) for x in items):
                raise TypeError(
                    f"Found blocks in this list that were not list [item]s!"
                )
            return DisplayList(numbered=True, items=cast(List[DisplayListItem], items))

        return enumerate_builder

    @dictify_pure_property
    def itemize(self) -> BlockScopeBuilder:
        @block_scope_builder
        def itemize_builder(contents: BlockScope) -> Block:
            items = list(contents)
            if any(not isinstance(x, DisplayListItem) for x in items):
                raise TypeError(
                    f"Found blocks in this list that were not list [item]s!"
                )
            return DisplayList(numbered=False, items=cast(List[DisplayListItem], items))

        return itemize_builder

    @dictify_pure_property
    def item(self) -> BlockScopeBuilder:
        @block_scope_builder
        def item_builder(block_scope: BlockScope) -> Block:
            return DisplayListItem(block_scope)

        return item_builder


class MarkdownUrlPlugin(RendererPlugin):
    def _inline_handlers(self) -> Iterable[CustomRenderFunc]:
        return ((NamedUrl, self._render_url),)

    def _render_url(self, renderer: Renderer, url: NamedUrl) -> str:
        if url.name is None:
            # Set the "name" of the URL to the text of the URL - escaped so it can be read as normal markdown
            escaped_url_name = renderer.render_unescapedtext(UnescapedText(url.url))
        else:
            escaped_url_name = renderer.render_inlinescope(url.name)
        return f"[{escaped_url_name}]({url.url})"

    def url(self, url: str, name: Optional[str] = None) -> Inline:
        return NamedUrl(
            url, name=InlineScope([UnescapedText(name)]) if name is not None else None
        )



from typing import Any, Callable, List, Optional

from turnip_text import (
    Block,
    BlockScope,
    BlockScopeBuilder,
    Inline,
    InlineScope,
    InlineScopeBuilder,
    RawScopeBuilder,
)


class block_scope_builder(BlockScopeBuilder):
    """
    Decorator which allows functions-returning-functions to fit the BlockScopeBuilder typeclass.

    e.g. one could define a function
    ```python
    def block(name=""):
        @block_scope_builder
        def inner(items: BlockScope) -> Block:
            return items
        return inner
    ```
    which allows turnip-text as so:
    ```!text
    [block(name="greg")]{
    The contents of greg
    }
    ```
    """

    func: Callable[[BlockScope], Optional[Block]]

    def __init__(self, func: Callable[[BlockScope], Optional[Block]]) -> None:
        self.func = func

    def build_from_blocks(self, b: BlockScope) -> Optional[Block]:
        return self.func(b)


class inline_scope_builder(InlineScopeBuilder):
    """
    Decorator which ensures functions fit the InlineScopeBuilder typeclass

    e.g. one could define a function
    ```python
    def inline(postfix = ""):
        @inline_scope_builder
        def inner(items: InlineScope) -> Inline:
            return InlineScope(list(items) + [postfix])
        return inner
    ```
    which allows turnip-text as so:
    ```!text
    [inline("!")]{surprise}
    ```
    """

    func: Callable[[InlineScope], Inline]

    def __init__(self, func: Callable[[InlineScope], Inline]) -> None:
        self.func = func

    def build_from_inlines(self, inls: InlineScope) -> Inline:
        return self.func(inls)


class raw_scope_builder(RawScopeBuilder):
    """
    Decorator which allows functions to fit the RawScopeBuilder typeclass.

    e.g. one could define a function
    ```python
    def math(name=""):
        @raw_scope_builder
        def inner(raw_text: str) -> Inline:
            return ...
        return inner
    ```
    which allows turnip-text as so:
    ```!text
    [math()]#{\sin\(x\)}#
    ```
    """

    func: Callable[[str], Inline]

    def __init__(self, func: Callable[[str], Inline]) -> None:
        self.func = func

    def build_from_inlines(self, raw: str) -> Inline:
        return self.func(raw)



import argparse
import json
from pathlib import Path

from turnip_text import *
from turnip_text.renderers.latex import LatexRenderer
from turnip_text.renderers.latex.plugins import (
    LatexCitationPlugin,
    LatexFootnotePlugin,
    LatexFormatPlugin,
    LatexListPlugin,
    LatexSectionPlugin,
    LatexUrlPlugin,
)
from turnip_text.renderers.markdown.base import MarkdownRenderer
from turnip_text.renderers.markdown.plugins import (
    MarkdownCitationAsFootnotePlugin,
    MarkdownCitationAsHTMLPlugin,
    MarkdownFootnotePlugin,
    MarkdownFormatPlugin,
    MarkdownListPlugin,
    MarkdownSectionPlugin,
    MarkdownUrlPlugin,
)


class CustomEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (BlockScope, InlineScope, Paragraph, Sentence)):
            return list(o)
        if isinstance(o, UnescapedText):
            return o.text
        if hasattr(o, "__dict__"):
            d = vars(o)
            d["str"] = str(o)
            return d
        return str(o)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-olatex", type=str)
    parser.add_argument("-omd", type=str)
    args = parser.parse_args()

    r_latex = LatexRenderer(
        [
            LatexCitationPlugin(),
            LatexFootnotePlugin(),
            LatexSectionPlugin(),
            LatexFormatPlugin(),
            LatexListPlugin(),
            LatexUrlPlugin(),
        ]
    )
    r_md = MarkdownRenderer(
        [
            MarkdownCitationAsHTMLPlugin(),
            MarkdownFootnotePlugin(),
            MarkdownSectionPlugin(),
            MarkdownFormatPlugin(),
            MarkdownListPlugin(),
            MarkdownUrlPlugin(),
        ]
    )
    r_md.request_postamble_order(
        [
            MarkdownFootnotePlugin._MARKDOWN_FOOTNOTE_POSTAMBLE_ID,
            MarkdownCitationAsHTMLPlugin._BIBLIOGRAPHY_POSTAMBLE_ID,
        ]
    )

    doc_block = r_latex.parse_file(Path("./examples/phdprop.ttxt"))
    rendered_latex = r_latex.render_doc(doc_block)
    if args.olatex:
        with open(args.olatex, "w") as f:
            f.write(rendered_latex)
    else:
        print(rendered_latex)

    doc_block = r_md.parse_file(Path("./examples/phdprop.ttxt"))
    rendered_markdown = r_md.render_doc(doc_block)
    if args.omd:
        with open(args.omd, "w") as f:
            f.write(rendered_markdown)
    else:
        print(rendered_markdown)

    # print(json.dumps(doc_block, indent=4, cls=CustomEncoder))



from dataclasses import dataclass
from typing import List

from turnip_text import Inline, UnescapedText
from turnip_text.renderers import Renderer, RendererPlugin


@dataclass(frozen=True)
class RawLatex(Inline):
    text: str


class LatexRenderer(Renderer):
    PARAGRAPH_SEP = "\n\n"
    SENTENCE_SEP = "\n"

    def __init__(self, plugins: List[RendererPlugin]) -> None:
        super().__init__(plugins)

        self.inline_handlers.push_association(
            (RawLatex, lambda _, raw: self.render_raw_latex(raw))
        )

    def render_raw_latex(self, r: RawLatex) -> str:
        return r.text

    def render_unescapedtext(self, t: UnescapedText) -> str:
        # note - right now this assumes we're using a unicode-compatible setup and thus don't need to escape unicode characters.
        # note - order is important here because the subsitutions may introduce more special characters.
        # e.g. if the backslash replacement applied later, the backslash from escaping "%" would be replaced with \textbackslash
        ascii_map = {
            "\\": "\\textbackslash{}",
            "%": "\%",
            "$": "\$",
            "{": "\{",
            "}": "\}",
            "_": "\_",
            "#": "\#",
            "&": "\&",
            "~": "\~{}",
        }
        data = t.text
        for c, replace_with in ascii_map.items():
            data = data.replace(c, replace_with)
        return data
